import argparse
import asyncio
from collections import namedtuple
from multiprocessing import freeze_support
import subprocess

from aiozmq import rpc

from pomoxis.align.minimap import MiniMapServe

from pomoxis import set_wakeup
from pomoxis.common import util

import logging
logger = logging.getLogger(__name__)


@asyncio.coroutine
def align_server(index, port, aligner, opts=''):
    bind = 'tcp://127.0.0.1:{}'.format(port)
    if aligner == 'minimap':
        server = yield from rpc.serve_rpc(
            MiniMapServe(index[0], map_opts=opts), bind=bind
        )
    else:
        raise ValueError("Unknown aligner '{}'.".format(aligner))
    return server


@asyncio.coroutine
def align_client(port):
    client = yield from rpc.connect_rpc(
        connect='tcp://127.0.0.1:{}'.format(port),
    )
    return client


class AlignClient(object):
    def __init__(self, port):
        """A synchronous alignment client.

        :param port: RPC server port.
        """
        self.port = port

    @asyncio.coroutine
    def _run_align(self, sequence):
        client = yield from align_client(self.port)
        results = yield from client.call.align(sequence)
        return results

    def align(self, sequence, loop=None):
        """Align a given sequence."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(self._run_align(sequence))
        loop.close()
        return result


def serve(args):
    @asyncio.coroutine
    def clean_up():
        # Doesn't seem right to have a client clean up the server
        logger.info('Creating cleanup client')
        client = yield from align_client(args.port)
        yield from client.call.clean_index()

    @asyncio.coroutine
    def _run():
        set_wakeup()
        server = yield from align_server(
            args.index, args.port, args.aligner, opts=args.opts
        )
        logger.info('Alignment server running, awaiting requests...')
        yield from server.wait_closed()

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run())
    except KeyboardInterrupt:
        logger.info('Shutting down server...')
        loop.run_until_complete(clean_up())
        logger.info('Server shut down.')


def send(args):
    client = AlignClient(args.port)
    for seq in args.sequences:
        print(client.align(seq))


def to_dict(string):
   """Convert a string whitespace separate key:value pairs to a dict."""
   return {k:v for k,v in (x.split(":") for x in string.split())}


def get_parser():
    parser = argparse.ArgumentParser('Alignment server/client.')
    subparsers = parser.add_subparsers(title='subcommands', description='valid commands', help='additional help', dest='command')
    subparsers.required = True

    sparser = subparsers.add_parser('server', help='Launch alignment server.')
    sparser.set_defaults(func=serve)
    sparser.add_argument('port', type=int, help='Port on which to serve.')
    sparser.add_argument('index', nargs='+', help='Filename path prefix for index files.')
    sparser.add_argument('--aligner', choices=('minimap'), default='minimap', help='Choice of aligner.')
    sparser.add_argument('--opts', default=dict(), type=to_dict, help="Alignment options as 'key:value key:value ...'.")

    sparser = subparsers.add_parser('client', help='Test client.')
    sparser.set_defaults(func=send)
    sparser.add_argument('port', type=int, help='Port on which to serve.')
    sparser.add_argument('sequences', nargs='+', help='Base sequences to align.')

    return parser


def main():
    freeze_support()
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    args = get_parser().parse_args()
    args.func(args)





import argparse
import sys
import os
import shutil
import copy
import platform
from collections import Counter
from multiprocessing import freeze_support
from timeit import default_timer as now

import asyncio
from aiozmq import rpc
from concurrent.futures import ProcessPoolExecutor as PoolExecutor
from functools import partial

import numpy as np
import msgpack

from fast5_research.fast5_bulk import BulkFast5

from pomoxis import set_wakeup

import logging
logger = logging.getLogger(__name__)


class Fast5Data(np.ndarray):

    def __new__(cls, input_array, info=None, start=None, end=None, sample_rate=6000):
        """Helper class for transmitting raw and event data over RPC requests.

        :param input_array: numpy array to encode.
        :param info: metadata.
        :param start: start time of data.
        :param end: end time of data.
        :param sample_rate: sampling rate of source data.
        """
        obj = np.asarray(input_array).view(cls)
        obj.info = info
        obj.start = start
        obj.end = end
        obj.sample_rate = sample_rate
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.info = getattr(obj, 'info', None)
        self.start = getattr(obj, 'start', None)
        self.end = getattr(obj, 'end', None)
        self.sample_rate = getattr(obj, 'sample_rate', None)

    @staticmethod
    def encode(obj):
        """Serialize data to send over RPC."""
        data = {
            b'shape': obj.shape,
            b'info': obj.info, 
            b'start': obj.start,
            b'end': obj.end,
            b'sample_rate': obj.sample_rate, 
            b'data': obj.tobytes(),
            b'type': obj.dtype.descr,
            b'kind': obj.dtype.kind
        }
        return data

    @staticmethod
    def decode(obj):
        """Deserialize data from RPC request."""
        if obj[b'kind'] == b'V':
            descr = [(x.decode('utf-8'), y.decode('utf-8')) for x, y in obj[b'type']]
        else:
            descr = obj[b'type'][0][1].decode('utf-8')
        data = Fast5Data(
            np.fromstring(
                obj[b'data'],
                dtype=np.dtype(descr)
            ).reshape(obj[b'shape']),
            info=obj[b'info'].decode('utf-8'),
            start=obj[b'start'],
            end=obj[b'end'],
            sample_rate=obj[b'sample_rate'],
        )
        return data


translation_table = {
    0: (Fast5Data,
        lambda value: msgpack.packb(value, default=value.encode),
        lambda binary: msgpack.unpackb(binary, object_hook=Fast5Data.decode)
    ),
}


class ReplayChannel(rpc.AttrHandler):

    def __init__(self, fast5, channel, *args, good_class='strand', time_warp=1, **kwargs):
        """An RPC service for replaying a channel from a .fast5 file.

        :param fast5: input filename.
        :param channel: channel to simulate.
        :param good_class: read classification name of desirable reads.
        :param time_warp: time multiplier for playback speed.

        ..note:: `args` and `kwargs` are passed to `aiozmq.rpc.AttrHandler`.
        
        """
        super().__init__(*args, **kwargs)
        self.fast5 = fast5
        self.channel = channel
        self.good_class = good_class
        self.time_warp = time_warp
        self.logger = logging.getLogger('Replay Channel {}'.format(channel.zfill(4)))

        with BulkFast5(self.fast5) as fh:
            self.sample_rate = fh.sample_rate
            self.time_base = 1.0 / self.sample_rate
            self.reads = [x for x in fh.get_reads(self.channel)]
            self.read_starts = np.ascontiguousarray([x['read_start'] for x in self.reads])

        self.reset_time()

    @rpc.method
    def reset_time(self):
        """Reset internal time to zero."""
        self.start_time = now()
        self.sample_offset = 0
        self._current_event = 0
        self._current_read = 0

    @property
    def current_sample(self):
        """Current simulated sample (the simulation time)."""
        return int((now() - self.start_time) * self.sample_rate * self.time_warp + self.sample_offset)


    @property
    def experiment_time(self):
        """Current simulated time."""
        return self.current_sample * self.time_base 

    @property
    def time_saved(self):
        """Difference between current real time and simulated time. (How far
        we've fast-forwarded the channel.)

        """
        return self.sample_offset * self.time_base

    @property
    def current_event(self):
        """Index of the current event."""
        prev = self._current_event
        with BulkFast5(self.fast5) as fh:
            event_path = fh.__event_data__.format(self.channel)
            new = np.searchsorted(fh[event_path][prev:]['start'], self.current_sample)
        self._current_event = int(new + prev)
        return self._current_event


    @property
    def current_read(self):
        """Index of the current read."""
        search_start = max(0, self._current_read - 2)
        self._current_read = search_start - 1 + np.searchsorted(self.read_starts[search_start:], self.current_sample)
        return self._current_read


    @property
    def good_reads_to_now(self):
        """Reads with classification equal to `good_class` until current time."""
        return [x for x in self.reads[:self._current_read + 1] if x['classification'] == self.good_class]
        
    @property
    def total_good_reads(self):
        """Total number of good reads seen up to current simulation time."""
        return len(self.good_reads_to_now)

    @property
    def cumulative_good_read_time_to_now(self):
        """Total duration of reads with classification equal to `good_class` until current time."""
        return sum(read['read_length'] for read in self.good_reads_to_now) * self.time_base


    @rpc.method
    def get_events(self, n_events=400):
        """Return events from the start of the current read.

        :param nevents: maximum number of events to return.

        :returns: Serialized event data, see :class:`Fast5Data`.
        """
        self.logger.debug("Request for events at {}".format(self.current_sample))
        read = self.reads[self.current_read]
        if read['classification'] != self.good_class:
            return None
        else:
            start = int(read['event_index_start'])
            end = min(self.current_event, start + n_events)
            self.logger.debug("Fetching events [{}, {}] for read {} starting at {}. Current sample is {}.".format(
                start, end,
                self.current_read, int(self.sample_offset + self.reads[self.current_read]['read_start']),
                self.current_sample
            ))
            with BulkFast5(self.fast5) as fh:
                events = fh.get_events(self.channel, event_indices=[start, end])
                return Fast5Data(
                    events, info=read['read_id'].decode('utf-8'),
                    start=int(self.sample_offset + read['read_start']),
                    end=int(self.sample_offset + read['read_start'] + read['read_length'])
                )

    @rpc.method
    def get_raw(self, seconds=1, delay=0.5):
        """Return events from the start of the current read.

        :param nevents: maximum number of events to return

        :returns: Serialized raw data, see :class:`Fast5Data`.
        """
        self.logger.debug("Request for raw at {}".format(self.current_sample))
        read = self.reads[self.current_read]
        if read['classification'] != self.good_class:
            return None
        else:
            start = int(read['read_start'] + self.sample_rate * delay)
            end = int(min(self.current_sample, start + self.sample_rate * seconds))
            if end <= start:
                return None
            self.logger.debug("Fetching raw [{}, {}] for read {} starting at {}. Current sample is {}.".format(
                start, end,
                self.current_read, int(self.sample_offset + self.reads[self.current_read]['read_start']),
                self.current_sample
            ))
            with BulkFast5(self.fast5) as fh:
                raw = fh.get_raw(self.channel, raw_indices=[start, end])
                return Fast5Data(
                    raw, info=read['read_id'].decode('utf-8'),
                    start=int(self.sample_offset + read['read_start']),
                    end=int(self.sample_offset + read['read_start'] + read['read_length'])
                )

    @rpc.method
    def unblock(self, read_id, read_end):
        """Fast forward to next read, returns new current sample. If the
        given read has already ended, no action is taken.

        :param read_id: read_id of read to unblock.
        :param read_end: expiration time for request, used only for logging.
        """
        cur_read = self.reads[self.current_read]
        if cur_read['read_id'].decode('utf-8') == read_id:
            next_read = self.reads[self.current_read + 1]
            jump = int(next_read['read_start'] - self.current_sample)
            self.sample_offset += jump
            self.logger.debug("Unblocking read {}. Skipping ahead {} samples to sample {}.".format(
                read_id, jump, self.current_sample
            ))
            return self.current_sample, True
        else:
            self.logger.info("Unblock received too late for read {}. Received at {}, read ended at {}".format(
                read_id, self.current_sample, read_end
            ))
            return self.current_sample, False


class ReplayFast5(rpc.AttrHandler):
    def __init__(self, fast5, channels, *args, good_class='strand', time_warp=1, **kwargs):
        """Replay multiple channels from a .fast5 file.

        :param fast5: input filename.
        :param channel: list of channels to simulate.
        :param good_class: read classification name of desirable reads.
        :param time_warp: time multiplier for playback speed.

        ..note:: `args` and `kwargs` are passed to `aiozmq.rpc.AttrHandler`.

        """
        super().__init__(*args, **kwargs)
        self.fast5 = fast5
        self.channels = channels
        self.good_class = good_class
        self.time_warp = time_warp
        self.logger = logging.getLogger('Replay Fast5')
        if self.time_warp < 1:
            self.logger.warn("time_warp set to slow down time, behaviour may be undefined.")

        self.replay_channels = {
            channel:ReplayChannel(self.fast5, channel, good_class=self.good_class, time_warp=self.time_warp)
            for channel in self.channels
        }

        for chan in self.replay_channels.values():
            chan.reset_time()

    @rpc.method
    def time_saved(self):
        """Difference between current real time and simulated time. (How far
        we've fast-forwarded the channel.)

        """
        return sum(chan.time_saved for chan in self.replay_channels.values())

    @rpc.method
    def total_good_reads(self):
        """Total number of good reads seen up to current simulation time."""
        return sum(chan.total_good_reads for chan in self.replay_channels.values())

    @rpc.method
    def cumulative_good_read_time(self):
        """Total duration of reads with classification equal to `good_class` until current time."""
        return sum(chan.cumulative_good_read_time_to_now for chan in self.replay_channels.values())

    @rpc.method
    def get_events(self, channel, n_events=400):
        """Return events from the start of the current read.

        :param nevents: maximum number of events to return.
        :param channel: channel for which to get events.

        :returns: Serialized event data, see :class:`Fast5Data`.
        """
        return self.replay_channels[channel].get_events(n_events=n_events)

    @rpc.method
    def get_raw(self, channel, seconds=0.5):
        """Return raw data from the start of the current read.

        :param channel: channel for which to get raw data.

        :returns: Serialized raw data, see :class:`Fast5Data`.
        """
        return self.replay_channels[channel].get_raw(seconds=seconds)

    @rpc.method
    def unblock(self, channel, read_id, read_end):
        """Fast forward to next read, returns new current sample. If the
        given read has already ended, no action is taken.

        :param channel: channel to unblock.
        :param read_id: read_id of read to unblock.
        :param read_end: expiration time for request, used only for logging.
        """
        return self.replay_channels[channel].unblock(read_id, read_end)


@asyncio.coroutine
def replay_server(fast5, channels, port, good_class, time_warp=1):
    """Create an RPC server to replay data from a .fast5 file.

    :param fast5: input .fast5 file for simulation.
    :param channels: list of channels to simulate.
    :param port: port on which to listen for clients.
    :param good_class: read classification name of desirable reads.
    
    :returns: instance of :class:`ReplayFast5`.
    """
    server = yield from rpc.serve_rpc(
        ReplayFast5(fast5, channels, good_class=good_class, time_warp=time_warp),
        bind='tcp://127.0.0.1:{}'.format(port),
        translation_table=translation_table,
    )
    return server


@asyncio.coroutine
def replay_client(port):
    """Create an RPC client to request raw/event data and send unblock requests.

    :param port: server port.
    """
    client = yield from rpc.connect_rpc(
        connect='tcp://127.0.0.1:{}'.format(port),
        translation_table=translation_table
    )
    return client


def read_until_demo(fast5, channels, port=5555):
    """Simple demo of read until server and client application.

    :param fast5: input .fast5 file.
    :param channels: list of channels to simulate.
    :param port: port on which to run server and cliant.
    """
    logger = logging.getLogger('ReadUntil App')
    good_class = 'strand'
    time_warp=2
    event_loop = asyncio.get_event_loop()
    set_wakeup()
    # Setup replay service
    event_loop.create_task(replay_server(
        fast5, channels, port, good_class, time_warp=time_warp
    ))

    @asyncio.coroutine
    def read_until(port):
        client = yield replay_client(port)
        counter = 0
        while True:
            #This can be mostly rewritten for any real app
            for channel in channels:
                read_block = yield from client.call.get_events(channel)
                if read_block is None:
                    logger.debug("Channel not in '{}' classification".format(good_class))
                else:
                    logger.info("Analysing {} events".format(len(read_block)))
                    yield from asyncio.sleep(0.5/time_warp)
                    counter += 1
                    # We decided to unblock
                    if counter % 10:
                        # We took too a long time
                        yield from asyncio.sleep(5/time_warp)
                    client.call.unblock(channel, read_block.info, read_block.end)
            yield from asyncio.sleep(5/time_warp)
    event_loop.create_task(read_until(port))

    try:
        event_loop.run_forever()
    except KeyboardInterrupt:
        pass


class ExpandRanges(argparse.Action):
    """Translate a str like 1,2,3-5,40 to [1,2,3,4,5,40]"""
    def __call__(self, parser, namespace, values, option_string=None):
        import re
        elts = []
        for item in values.replace(' ', '').split(','):
            mo = re.search(r'(\d+)-(\d+)', item)
            if mo is not None:
                rng = [int(x) for x in mo.groups()]
                elts.extend(list(range(rng[0], rng[1] + 1)))
            else:
                elts.append(int(item))
        setattr(namespace, self.dest, elts)

if __name__ == '__main__':
    freeze_support()
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    parser = argparse.ArgumentParser('Simple zmq server and client demo.')
    parser.add_argument('fast5', help='Input fast5.')
    parser.add_argument('channels', action=ExpandRanges, help='Fast5 channel for source data.')
    args = parser.parse_args()
    read_until_demo(args.fast5, [str(x) for x in args.channels])



import argparse
import pysam
import sys
from Bio import SeqIO
from Bio.Seq import Seq


def _gen_seqs(bam):
    """Yield SeqIO records of the reference sequence each query is aligned to."""
    for read in bam:
        ref_seq = read.get_reference_sequence().upper()
        name = '{}_{}'.format(read.reference_name, read.query_name)
        yield SeqIO.SeqRecord(id=name, seq=Seq(ref_seq))


def main():
    parser = argparse.ArgumentParser('Extract reference sequence that queries are aligned to')
    parser.add_argument('bam', help='input bam file, MD tag must be set (mini_align -m).')

    args = parser.parse_args()

    with pysam.AlignmentFile(args.bam) as bam:
        SeqIO.write(_gen_seqs(bam), sys.stdout, 'fasta')


if __name__ == '__main__':
    main()



import asyncio
import os
from watchdog.observers import Observer
from watchdog.events import RegexMatchingEventHandler
from watchdog.utils import has_attribute, unicode_paths
import logging

logger = logging.getLogger(__name__)

EVENT_TYPE_MOVED = 'moved'
EVENT_TYPE_DELETED = 'deleted'
EVENT_TYPE_CREATED = 'created'
EVENT_TYPE_MODIFIED = 'modified'


def wait_for_file(fname):
    """Block until a filesize remains constant."""
    # watchdog cannot tell us when a newly created file has finished being
    #    written. Let's just wait until the filesize stops changing
    size = None
    while True:
        try:
            newsize = os.path.getsize(fname)
        except:
            newsize = None
        else:
            if newsize is not None and size == newsize:
                break
        print(size, newsize, fname)
        size = newsize

class AIORegexMatchingEventHandler(RegexMatchingEventHandler):
    def __init__(self, callback, loop=None, **kwargs):
        """asyncio compatible minimal regex matching event
        handler for watchdog.

        :param callback: function to apply to filenames.
        :param loop: ayncio-like event loop.
        """

        RegexMatchingEventHandler.__init__(self, **kwargs)
        self._loop = loop or asyncio.get_event_loop()
        self.callback = callback

    @asyncio.coroutine
    def _process_file(self, event):
        """Process an event when a file is created (or moved).

        :param event: watchdog event.

        :returns: result of applying `callback` to watched file.
        """
        if event.event_type == EVENT_TYPE_CREATED:
            fname = event.src_path
        else:
            fname = event.dest_path
        # need to wait for file to be closed
        wait_for_file(fname)
        result = yield from self.callback(fname)
        return result

    def dispatch(self, event):
        """Dispatch an event after filtering. We handle
        creation and move events only.
        
        :param event: watchdog event.

        :returns: None
        """
        if event.event_type not in (EVENT_TYPE_CREATED, EVENT_TYPE_MOVED):
            return
        if self.ignore_directories and event.is_directory:
            return

        paths = []
        if has_attribute(event, 'dest_path'):
            paths.append(unicode_paths.decode(event.dest_path))
        if event.src_path:
            paths.append(unicode_paths.decode(event.src_path))

        if any(r.match(p) for r in self.ignore_regexes for p in paths):
            return

        if any(r.match(p) for r in self.regexes for p in paths):
            self._loop.call_soon_threadsafe(asyncio.async, self._process_file(event))


class Watcher(object):
    def __init__(self, path, event_handler, recursive=False):
        """Wrapper around common watchdog idiom.

        :param path: path to watch for new files.
        :param event_handler: subclass of watchdog.events.FileSystemEventHandler.
        :param recursive: watch path recursively?
        """
        self.observer = Observer()
        self.observer.schedule(event_handler, path, recursive)

    def start(self):
        """Start observing path."""
        self.observer.start()

    def stop(self):
        """Stop observing path."""
        self.observer.stop()
        self.observer.join()


@asyncio.coroutine
def watch_path(path, callback, recursive=False, regexes=['.*\.fast5$']):
    """Watch a filepath indefinitely for new files, applying callback to files.

    :param path: path to watch.
    :param callback: callback to apply to newly found files.
    :param recursive: recursively watch path?
    :param regexes: filter files by applying regex.
    """
    handler = AIORegexMatchingEventHandler(callback=callback, regexes=regexes)
    watch = Watcher(path, event_handler=handler, recursive=recursive)
    logging.info('Starting to watch {} for new files matching {}.'.format(path, regexes))
    watch.start()



"""
Tabulate some simple alignment stats from sam.
"""

import argparse
import sys
import numpy as np
import pandas as pd
import warnings
from collections import OrderedDict

parser = argparse.ArgumentParser(
        description="""Summarise output of `stats_from_bam`.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('-i', '--input', type=argparse.FileType('r'), default=sys.stdin)
parser.add_argument('-o', '--output', type=argparse.FileType('w'), default=sys.stderr,
                    help='Output summary to file instead of stderr.')
parser.add_argument('-p', '--percentiles', type=int, nargs='+', default=(10,50,90),
                    help='Percentiles for summary.')
parser.add_argument('-pr', '--per_reference', action='store_true',
                    help='Also output a summary for each reference.')

def qscore(d):
    with warnings.catch_warnings():
        # some values might be zero
        warnings.simplefilter("ignore")
        q = -10 * np.log10(d)
    return q

def summarise_stats(d, percentiles=(10, 50, 90)):

    s = []
    def summarise_stat(name, f):
        # per alignment errors from which to calculate metrics, bounds etc.
        r = OrderedDict()
        r['name'] = name
        r['mean'] = f[0](d)
        vals = f[1](d)
        for p, pval in zip(percentiles, np.percentile(vals, percentiles)):
            r['q{}'.format(p)] = pval
        return r

    # create functions to get overall length-weighted averages
    # and per-alignment errors from which to calculate percentiles
    f = OrderedDict()
    f['err_ont'] = (lambda d: np.sum(d['sub'] + d['ins'] + d['del']) / np.sum(d['length']), # weighted
                    lambda d: (d['sub'] + d['ins'] + d['del'])/ d['length']) # per alignment
    f['err_bal'] = (lambda d: np.sum(d['sub'] + d['ins'] + d['del']) / np.sum(d['rend'] - d['rstart']),
                    lambda d: (d['sub'] + d['ins'] + d['del'])/ (d['rend'] - d['rstart']))
    f['iden'] = (lambda d: np.sum(d['sub']) / (np.sum(d['rend'] - d['rstart'] - d['del'])),
                 lambda d: d['sub'] / (d['rend'] - d['rstart'] - d['del']))
    f['del'] = (lambda d: np.sum(d['del']) / (np.sum(d['rend'] - d['rstart'])),
               lambda d: d['del'] / (d['rend'] - d['rstart']))
    f['ins'] = (lambda d: np.sum(d['ins']) / (np.sum(d['rend'] - d['rstart'])),
                lambda d: d['ins'] / (d['rend'] - d['rstart']))

    for name, f in f.items():
        s.append(summarise_stat(name, f))

    return s


def main(arguments=None):
    args = parser.parse_args(arguments)

    stats = pd.read_table(args.input)

    if len(stats) == 0:
        raise ValueError('No alignments stats found.')


    def output_summary(stats, prefix=''):
        to_str_opts = {'buf': args.output, 'col_space': 8, 'index':False, 'justify': 'center'}
        errors = pd.DataFrame(summarise_stats(stats, percentiles=args.percentiles))
        pcnt_formatter = {}
        for col in errors.columns:
            if np.issubdtype(errors[col].dtype, np.number):
                pcnt_formatter[col] = '{:5.3%}'.format
            else:
                pcnt_formatter[col] = '{}'.format
        args.output.write('\n\n# {} Percentage Errors\n'.format(prefix))
        errors.to_string(formatters=pcnt_formatter, **to_str_opts)

        args.output.write('\n\n# {} Q Scores\n'.format(prefix))
        for col in errors.columns:
            if np.issubdtype(errors[col].dtype, np.number):
                errors[col] = qscore(errors[col])
        errors.to_string(float_format='%5.2f', **to_str_opts)

    output_summary(stats)

    args.output.write('\n\n# Ref Coverage\n')
    for ref, grp in stats.groupby('ref'):
        args.output.write('{} {:5.3f}\n'.format(ref, grp['ref_coverage'].sum()))

    if args.per_reference:
        for ref, grp in stats.groupby('ref'):
            output_summary(grp, prefix=ref)


if __name__ == '__main__':
    main()



import asyncio
from collections import namedtuple

from aiozmq import rpc

import mappy

import logging
logger = logging.getLogger(__name__)


class MiniMapServe(rpc.AttrHandler):

    def __init__(self, index, *args, map_opts={'preset':'map-ont'}, **kwargs):
        """bwa mem alignment server implementation using python binding.

        :param index: bwa index base path, or list thereof.
        :param map_opts: command line options for minimap

        """
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger('MiniMapServe')
        self.index = index
        self.map_opts = map_opts

        self.aligner = None
        self.aligner = mappy.Aligner(self.index, **map_opts)
        self.logger.info('Minimap service started.')

    def _clean_index(self):
        self.logger.info('Cleaning alignment proxy.')
        self.aligner = None

    def __del__(self):
            self._clean_index()

    @rpc.method
    def clean_index(self):
        """Destroy the aligner object, which will cleanup the index in memory."""
        self._clean_index()

    @rpc.method
    @asyncio.coroutine
    def align(self, sequence):
        """Align a base sequence.

        :param sequence: sequence to align.

        :returns: the output of bwa mem call.
        """
        if self.aligner is None:
            self.aligner = mappy.Aligner(self.index, **map_opts)
        self.logger.debug("Aligning sequence of length {}.".format(len(sequence)))
        # create named tuples from the results
        results = []
        for r in self.aligner.map(sequence):
             results.append((
                 r.q_st, r.q_en, '+' if r.strand > 0 else '-',
                 r.ctg, r.ctg_len, r.r_st, r.r_en,
                 r.mlen, r.blen, r.mapq, r.is_primary,
                 r.cigar_str
             ))
        self.logger.info("Aligned sequence of {} bases with {} hits.".format(
            len(sequence), len(results)
        ))
        return results




import argparse
from concurrent.futures import ThreadPoolExecutor
import functools
import itertools
import logging

import numpy as np
from pysam import FastaFile

import scrappy

from pomoxis.common.bio import reverse_complement, shotgun_library


def worker(args, noise=None):
    model='rgrgr_r94'
    seq, ref, start, end, strand = args
    squiggle = scrappy.sequence_to_squiggle(
        seq, rescale=True).data(as_numpy=True, sloika=False)

    n = 1/np.sqrt(2)
    raw_data = np.concatenate([
        np.random.laplace(mean, n*stdv, int(dwell))
        for mean, stdv, dwell in squiggle
    ])
    if noise is not None:
        raw_data += np.random.normal(scale=noise, size=len(raw_data))

    raw = scrappy.RawTable(raw_data)
    raw.scale()
    post = scrappy.calc_post(raw, model, log=True)
    call, score, _ = scrappy.decode_post(post, model)
    return '>call_{}:{}-{}({}) seq_len={} call_len={} score={}\n{}'.format(
        ref, start, end, strand, len(seq), len(call), score, call)


def main():
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    parser = argparse.ArgumentParser(description='Simulate basecalls with scrappy.')

    parser.add_argument('fasta', help='Source sequence file.')
    parser.add_argument('ncalls', type=int, help='Number of basecalls to produce.')
    parser.add_argument('--mu', type=float, default=8000, help='mean fragment length.')
    parser.add_argument('--sigma', type=float, default=1000, help='stdv fragment length.')
    parser.add_argument('--noise', type=float, default=0.06, help='Additional Gaussian noise on signal.')
    parser.add_argument('--threads', type=int, default=None, help='number of worker threads.')

    args = parser.parse_args()

    regions = itertools.islice(shotgun_library(
        args.fasta, args.mu, args.sigma, direction=(1,-1)
    ), args.ncalls)

    _worker = functools.partial(worker, noise=args.noise)
    with ThreadPoolExecutor(max_workers=args.threads) as executor:
        for fasta in executor.map(_worker, regions):
            print(fasta)


if __name__ == '__main__':
    main()



import argparse
import pysam
from collections import Counter
from pomoxis.common.summary_from_stats import qscore
from pomoxis.common.stats_from_bam import stats_from_aligned_read


def get_errors(aln, ref_seq=None):
    seq = aln.query_sequence
    errors = {}
    insertions = ''
    pairs = aln.get_aligned_pairs(with_seq=True)
    if pairs[0][0] is None or pairs[0][1] is None:
        raise ValueError('It does not look like bam is trimmed to a common alignment window')
    for qp, rp, rb in pairs[::-1]:  # process pairs in reverse to easily accumulate insertions
        if qp is None:  # deletion
            errors[rp] = (rb, '-')
        elif rp is None:  # insertion
            insertions += seq[qp]
        # if we reach here, qp is not None and rp is not None
        elif len(insertions) > 0:
            # this also includes cases where the ref and query don't agree
            # e.g. ref   A-TGC
            #      query GTTGC  would emit, A -> GT
            errors[rp] = (rb.upper(), seq[qp] + insertions[::-1])
            insertions = ''
        elif seq[qp] != rb:  # mismatch
            errors[rp] = (rb.upper(), seq[qp])
        if ref_seq is not None and rp is not None:
            assert ref_seq[rp] == rb.upper()

    if ref_seq is not None:
        # check we can scuff up reference using errors and get back our query
        scuffed = scuff_ref(ref_seq, errors)
        if not scuffed == aln.query_sequence:
            raise ValueError('Scuffing up reference with errors did not recreate query')
    return errors


def count_errors(errors):
    counts = Counter()
    for rp, (ref, query) in errors.items():
        if query == '-':
            counts['del'] += 1
        elif ref != query[0]:
            counts['sub'] += 1
        if len(query) > 1:
            counts['ins'] += len(query) - 1
    return counts


def scuff_ref(ref_seq, errors):
    orig_seq = list(ref_seq)
    for rp, (rb, alt) in errors.items():
        assert orig_seq[rp] == rb
        if alt == '-':
            orig_seq[rp] = ''
        else:
            orig_seq[rp] = alt
    return ''.join(orig_seq)


def get_qscores(counts, ref_len):
    return {
        'Q(acc)': qscore(sum(counts.values()) / ref_len),
        'Q(iden)': qscore(counts['sub'] / (ref_len - counts['del'])),
        'Q(ins)': qscore(counts['ins'] / ref_len),
        'Q(del)': qscore(counts['del'] / ref_len),
    }


def main():
    parser = argparse.ArgumentParser('Get errors common to multiple assemblies aligned to ref.')
    parser.add_argument('bam', help='input bam file containing assemblies trimmed to a common alignment window')
    parser.add_argument('ref_fasta', help='reference fasta file of the reference over that alignment window')
    parser.add_argument('-o', '--output_prefix', default='common_errors',
                        help='Prefix for outputs.')

    args = parser.parse_args()

    bam = pysam.AlignmentFile(args.bam)
    if len(bam.references) > 1:
        raise ValueError('Bam should have just one reference')
    ref_lengths = dict(zip(bam.references, bam.lengths))

    ref_seq = pysam.FastaFile(args.ref_fasta).fetch(bam.references[0])

    # reads should already be trimmed to a common aligment start and end point
    reads = [r for r in bam]
    ref_end, ref_start = reads[0].reference_end, reads[0].reference_start
    ref_len = ref_end - ref_start
    if not (all([r.reference_end == ref_end for r in reads]) and
            all([r.reference_start == ref_start for r in reads])):
        raise ValueError('Alignments have not been trimmed to a common overlap window, try trim_alignments')

    # get errors in each read
    data = {}
    qscores = []
    for aln in reads:
        errors = get_errors(aln, ref_seq)
        counts = count_errors(errors)
        # check we got the same error counts as stats_from_aligned_read
        stats = stats_from_aligned_read(aln, list(ref_lengths.keys()), list(ref_lengths.values()))
        for k in counts.keys():
            if stats[k] != counts[k]:
                msg = "Error counts {} don't match those from the CIGAR str {}."
                raise ValueError(msg.format(counts, {k: stats[k] for k in counts.keys()}))
        qscores.append((aln.query_name, get_qscores(counts, ref_len)))
        data[aln.query_name] = errors

    # get intersection of errors
    names = list(data.keys())
    common_errors = set(data[names[0]].keys())  # set of reference positions
    for name in names[1:]:
        common_errors = common_errors.intersection(set(data[name].keys()))
    remaining_errors = {}
    # loop through common errors, checking ref is the same and retaining the
    # error with the shortest edit distance
    for rp in common_errors:
        ref = data[names[0]][rp][0]
        assert all([d[rp][0] == ref for d in data.values()])  # refs should be same
        alts = [d[rp][1] for d in data.values()]
        if len(set([len(alt) for alt in alts])) > 1:
            # we should take the best one
            alts = sorted(alts, key=lambda x: len(x))
            shortest = alts[0]
            others = alts[1:]
            if shortest == '-' and any([alt[0] == ref for alt in others]):
                # the alt with the insertion contained the ref,
                # and the alt with the deletion did not contain the insertion
                # so two wrongs make a right!
                continue
            else:
                remaining_errors[rp] = (ref, shortest)
        else:  # pick most common, arbitrary for equal numbers
            remaining_errors[rp] = (ref, Counter(alts).most_common()[0][0])

    # write fasta of ref scuffed with just common errors
    ref_scuffed = scuff_ref(ref_seq, remaining_errors)
    with open('{}.fasta'.format(args.output_prefix), 'w') as fh:
        fh.write('>{}\n'.format(args.output_prefix))
        fh.write(ref_scuffed)

    remaining_counts = count_errors(remaining_errors)
    qscores.append(('common_errors', get_qscores(remaining_counts, ref_len)))

    # print qscores of individual reads and overlapping errors
    cols = ['Q(acc)', 'Q(iden)', 'Q(del)', 'Q(ins)']
    with open('{}.txt'.format(args.output_prefix), 'w') as fh:
        fh.write('\t'.join(['name'] + cols) + '\n')
        for name, d in qscores:
            fh.write('\t'.join([name] + [str(d[c]) for c in cols]) + '\n')


if __name__ == '__main__':
    main()



import argparse
from collections import namedtuple, defaultdict
import itertools
import shutil
import sys

from Bio import SeqIO
import intervaltree
import numpy as np
from pysam import FastxFile

Region = namedtuple('Region', 'ref_name start end')
AlignPos = namedtuple('AlignPos', ('qpos', 'qbase', 'rpos', 'rbase'))


def chunks(iterable, n):
    """Generate fixed length chunks of an interable.

    :param iterable: input sequence.
    :param n: chunk size.
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def cat(files, output, chunks=1024*1024*10):
    """Concatenate a set of files.

    :param files: input filenames.
    :param output: output filenames.
    :param chunks: buffersize for filecopy.
    """
    with open(output, 'wb') as wfd:
        for f in files:
            with open(f, 'rb') as fd:
                shutil.copyfileobj(fd, wfd, chunks)


def split_fastx(fname, output, chunksize=10000):
    """Split records in a fasta/q into fixed lengths.

    :param fname: input filename.
    :param output: output filename.
    :param chunksize: (maximum) length of output records.
    """
    with open(output, 'w') as fout:
        with FastxFile(fname, persist=False) as fin:
            for rec in fin:
                name = rec.name
                seq = rec.sequence
                qual = rec.quality
                if rec.comment is None:
                    comment = 'chunk_length={}'.format(chunksize)
                else:
                    comment = '{} chunk_length={}'.format(rec.comment, chunksize)
                if qual is None:
                    for i, s in enumerate(chunks(seq, chunksize)):
                        chunk_name = '{}_chunk{}'.format(name, i)
                        fout.write(">{} {}\n{}\n".format(
                            chunk_name, comment, ''.join(s)))
                else:
                    for i, (s, q) in enumerate(zip(chunks(seq, chunksize), chunks(qual, chunksize))):
                        chunk_name = '{}_chunk{}'.format(name, i)
                        fout.write('@{} {}\n{}\n+\n{}\n'.format(
                            chunk_name, comment, ''.join(s), ''.join(q)))


def split_fastx_cmdline():
    """Split records in a fasta/q file into chunks of a maximum size."""
    fname, output, chunksize = sys.argv[1:]
    split_fastx(fname, output, int(chunksize))


def fast_convert():
    """Convert between fasta<->fastq."""

    parser = argparse.ArgumentParser(description='fast_convert -- Convert between fasta<->fastq.')
    parser.add_argument('convert', choices=['qq', 'aa', 'aq', 'qa'],
        help='Conversion code: from->to.')
    parser.add_argument('--discard_q', action='store_true',
        help='Discard quality information from fastq, use with --mock_q.')
    parser.add_argument('--mock_q', default=10, type=int,
        help='Mock quality value, valid for convert=aq|qq.')
    args = parser.parse_args()

    in_fmt = 'fastq'
    out_fmt= 'fasta'
    qflag = False # controls quality manipulation
    if args.convert == 'qq':
        out_fmt = 'fastq'
        if args.discard_q is not None:
            qflag = True
    elif args.convert == 'aa':
        in_fmt = 'fasta'
    elif args.convert == 'aq':
        in_fmt = 'fasta'
        out_fmt = 'fastq'
        qflag = True
    elif args.convert == 'qa':
        pass # default
    else:
        raise ValueError("convert must be 'qq', 'aq', 'qa', or 'aa'\n")

    if qflag:
        def fq_gen(io):
            for rec in SeqIO.parse(io, in_fmt):
                rec.letter_annotations["phred_quality"] = [args.mock_q] * len(rec)
                yield rec
        sys.stderr.write('Creating/ignoring quality information in input.\n')
        SeqIO.write(fq_gen(sys.stdin), sys.stdout, out_fmt)
    else:
        SeqIO.convert(sys.stdin, in_fmt, sys.stdout, out_fmt)


def extract_long_reads():
    """Filter fastq to longest reads."""

    parser = argparse.ArgumentParser(description='Extract longest reads from a fastq.')
    parser.add_argument('input',
        help='Input .fastq file.')
    parser.add_argument('output',
        help='Output .fastq file.')
    filt = parser.add_mutually_exclusive_group(required=True)
    filt.add_argument('--longest', default=None, type=float,
        help='Percentage of longest reads to partition.')
    filt.add_argument('--bases', default=None, type=int,
        help='Maximum number of bases (subject to at least one read.)')
    parser.add_argument('--others', default=None,
        help='Write all other reads to file.')
    args = parser.parse_args()

    sys.stderr.write('Loading reads...\n')
    fmt = 'fastq'
    try:
        record_dict = SeqIO.index(args.input, fmt)
    except:
        fmt = 'fasta'
        record_dict = SeqIO.index(args.input, fmt)
    sys.stderr.write('Format is {}.\n'.format(fmt))

    ids = list(record_dict.keys())
    lengths = np.fromiter(
        (len(record_dict[i]) for i in ids),
        dtype=int, count=len(ids)
    )
    sys.stderr.write('Sorting reads...\n')
    if args.bases is None:
        # partial sort will do fine here
        max_reads = int(len(ids) * (args.longest / 100))
        longest = np.argpartition(lengths, -max_reads)[-max_reads:]
    else:
        # need a full sort
        order = np.argsort(lengths)[::-1]
        cumsum = 0
        last = 1
        for i, j in enumerate(np.argsort(lengths), 1):
            cumsum += lengths[j]
            if cumsum > args.bases:
                break
            last = i
        longest = order[:last]

    SeqIO.write(
        (record_dict[ids[i]] for i in longest),
        args.output, fmt
    )

    if args.others is not None:
        longest = set(longest)
        SeqIO.write(
            (record_dict[ids[i]] for i in range(len(ids)) if i not in longest),
            args.others, fmt
        )


def parse_regions(regions, ref_lengths=None):
    """Parse region strings into `Region` objects.

    :param regions: iterable of str
    :param ref_lengths: {str ref_names: int ref_lengths}, if provided Region.end
        will default to the reference length instead of None.

    >>> parse_regions(['Ecoli'])[0]
    Region(ref_name='Ecoli', start=0, end=None)
    >>> parse_regions(['Ecoli:1000-2000'])[0]
    Region(ref_name='Ecoli', start=1000, end=2000)
    >>> parse_regions(['Ecoli:-1000'])[0]
    Region(ref_name='Ecoli', start=0, end=1000)
    >>> parse_regions(['Ecoli:500-'])[0]
    Region(ref_name='Ecoli', start=500, end=None)
    >>> parse_regions(['Ecoli'], ref_lengths={'Ecoli':4800000})[0]
    Region(ref_name='Ecoli', start=0, end=4800000)
    >>> parse_regions(['NC_000921.1:10000-20000'])[0]
    Region(ref_name='NC_000921.1', start=10000, end=20000)
    """
    def _get_end(end, ref_name):
        return end if end is not None else (
            ref_lengths[ref_name] if ref_lengths is not None else None
        )

    decoded = []
    for region in regions:
        if ':' not in region:
            ref_name, start, end = region, 0, None
        else:
            start, end = 0, None
            ref_name, bounds = region.split(':')
            if bounds[0] == '-':
                start = 0
                end = int(bounds[1:])
            elif bounds[-1] == '-':
                start = int(bounds[:-1])
                end = None
            else:
                start, end = [int(b) for b in bounds.split('-')]
        decoded.append(Region(ref_name, start, _get_end(end, ref_name)))
    return tuple(decoded)


class SeqLen(argparse.Action):
    """Parse a sequence length from str such as 4.8mb or from fastx."""
    def __call__(self, parser, namespace, values, option_string=None):
        seq_len = None
        try:
            seq_len = int(values)
        except:
            suffixes = {
                'kb': 10 ** 3,
                'mb': 10 ** 6,
                'gb': 10 ** 9,
            }
            for suffix, multiplier in suffixes.items():
                if suffix in values.lower():
                    seq_len = int(multiplier * float(values.lower().replace(suffix, '')))
                    break
            if seq_len is None:
                try:
                    seq_len = sum(get_seq_lens(values))
                except:
                    raise ValueError('Could not get sequence length from {}.'.format(values))

        setattr(namespace, self.dest, seq_len)


def get_seq_lens(fastx):
    """Get sequence lengths from fastx file"""
    return [len(r.sequence) for r in FastxFile(fastx)]


def coverage_from_fastx():
    parser = argparse.ArgumentParser('Estimate coverage from summed basecall and reference lengths')
    parser.add_argument('basecalls', help='input fastx file.')
    parser.add_argument('ref_len', action=SeqLen,
                        help='reference length (e.g. 4.8kb/mb/gb) or reference fastx from which to calculate length.')
    parser.add_argument('--coverage', type=int, help='Calculate fraction of reads required for this coverage.')
    parser.add_argument('--longest', action='store_true', default=False,
                        help='Use the longest reads when calculating fraction reads required for a given coverage.')

    args = parser.parse_args()

    basecall_lens = get_seq_lens(args.basecalls)
    total_basecalls_len = sum(basecall_lens)

    print('Total length of basecalls: {}'.format(total_basecalls_len))
    print('Total ref length: {}'.format(args.ref_len))
    print('Estimated coverage: {:.2f}'.format(total_basecalls_len / args.ref_len))

    if args.coverage is not None:
        if args.longest:
            basecall_lens.sort(reverse=True)
            cum_len = np.cumsum(basecall_lens)
            required_len = args.coverage * args.ref_len
            n_required = np.searchsorted(cum_len, required_len, side='right')
            msg = '% of longest reads required to achieve {}X coverage: {}'
            print(msg.format(args.coverage, 100 * n_required / len(cum_len)))
        else:
            msg = '% of randomly-selected reads required to achieve {}X coverage: {}'
            print(msg.format(args.coverage, 100 * args.coverage  * args.ref_len / total_basecalls_len))


def get_pairs(aln):
    """Return generator of pairs.

    :param aln: `pysam.AlignedSegment` object.
    :returns: generator of `AlignPos` objects.
    """
    seq = aln.query_sequence
    pairs = (AlignPos(qpos=qp,
                      qbase=seq[qp] if qp is not None else '-',
                      rpos=rp,
                      rbase=rb if rp is not None else '-'
                      )
             for qp, rp, rb in aln.get_aligned_pairs(with_seq=True)
             )
    return pairs


def get_trimmed_pairs(aln):
    """Trim aligned pairs to the alignment.

    :param aln: `pysam.AlignedSegment` object
    :yields pairs:
    """

    pairs = get_pairs(aln)
    for pair in itertools.dropwhile(lambda x: x.rpos is None or x.qpos is None, pairs):
        if (pair.rpos == aln.reference_end or pair.qpos == aln.query_alignment_end):
            break
        yield pair


def yield_from_bed(bedfile):
    with open(bedfile) as fh:
        for line in fh:
            split_line = line.split()
            if split_line[0] in {'browser', 'track'} or len(split_line) < 3:
                continue
            chrom = split_line[0]
            start = int(split_line[1])
            stop = int(split_line[2])
            yield chrom, start, stop


def intervaltrees_from_bed(path_to_bed):
    """Created dict of intervaltrees from a .bed file, indexed by chrom.

    :param path_to_bed: str, path to .bed file.
    :param chrom: str, chromosome name.
    :returns: { str chrom: `intervaltree.IntervalTree` obj }.
    """
    trees = defaultdict(intervaltree.IntervalTree)
    for chrom, start, stop in yield_from_bed(path_to_bed):
        trees[chrom].add(intervaltree.Interval(begin=start, end=stop))
    return trees



"""
Tabulate some simple alignment stats from sam.
"""

import argparse
from collections import Counter
from concurrent.futures import ProcessPoolExecutor
import functools
import itertools
from math import ceil
import sys

import pysam

from pomoxis.common.util import intervaltrees_from_bed

parser = argparse.ArgumentParser(
        description="""Parse a bamfile (from a stream) and output summary stats for each read.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('bam', type=str, help='Path to bam file.')
parser.add_argument('--bed', default=None, help='.bed file of reference regions to include.')
parser.add_argument('-m', '--min_length', type=int, default=None)
parser.add_argument('-a', '--all_alignments', action='store_true',
                    help='Include secondary and supplementary alignments.')
parser.add_argument('-o', '--output', type=argparse.FileType('w'), default=sys.stdout,
                    help='Output alignment stats to file instead of stdout.')
parser.add_argument('-s', '--summary', type=argparse.FileType('w'), default=sys.stderr,
                    help='Output summary to file instead of stderr.')
parser.add_argument('-t', '--threads', type=int, default=1,
                    help='Number of threads for parallel processing.')


def stats_from_aligned_read(read, references, lengths):
    """Create summary information for an aligned read

    :param read: :class:`pysam.AlignedSegment` object
    """

    tags = dict(read.tags)
    try:
        tags.get('NM')
    except:
        raise IOError("Read is missing required 'NM' tag. Try running 'samtools fillmd -S - ref.fa'.")

    name = read.qname
    counts, _ = read.get_cigar_stats()
    match = counts[0] # alignment match (can be a sequence match or mismatch)
    ins = counts[1]
    delt = counts[2]
    # NM is edit distance: NM = INS + DEL + SUB
    sub = tags['NM'] - ins - delt
    length = match + ins + delt
    iden = 100*float(match - sub)/match
    acc = 100 - 100*float(tags['NM'])/length

    read_length = read.infer_read_length()
    coverage = 100*float(read.query_alignment_length) / read_length
    direction = '-' if read.is_reverse else '+'

    results = {
        "name": name,
        "coverage": coverage,
        "direction": direction,
        "length": length,
        "read_length": read_length,
        "match": match,
        "ins": ins,
        "del": delt,
        "sub": sub,
        "iden": iden,
        "acc": acc,
        "qstart": read.query_alignment_start,
        "qend": read.query_alignment_end,
        "rstart": read.reference_start,
        "rend": read.reference_end,
        "ref": references[read.reference_id],
        "aligned_ref_len": read.reference_length,
        "ref_coverage": 100*float(read.reference_length) / lengths[read.reference_id],
    }

    return results


def masked_stats_from_aligned_read(read, references, lengths, tree):
    """Create summary information for an aligned read over regions in bed file.

    :param read: :class:`pysam.AlignedSegment` object
    """

    tags = dict(read.tags)
    try:
        tags.get('MD')
    except:
        raise IOError("Read is missing required 'MD' tag. Try running 'samtools callmd - ref.fa'.")

    correct, delt, ins, sub, aligned_ref_len, masked = 0, 0, 0, 0, 0, 0
    pairs = read.get_aligned_pairs(with_seq=True)
    qseq = read.query_sequence
    pos_is_none = lambda x: (x[1] is None or x[0] is None)
    pos = None
    insertions = []
    # TODO: refactor to use get_trimmed_pairs (as in catalogue_errors)?
    for qp, rp, rb in itertools.dropwhile(pos_is_none, pairs):
        if rp == read.reference_end or (qp == read.query_alignment_end):
            break
        pos = rp if rp is not None else pos
        if not tree.overlaps(pos) or (rp is None and not tree.overlaps(pos + 1)):
            # if rp is None, we are in an insertion, check if pos + 1 overlaps
            # (ref position of ins is arbitrary)
            # print('Skipping ref {}:{}'.format(read.reference_name, pos))
            masked += 1
            continue
        else:
            if rp is not None:  # we don't have an insertion
                aligned_ref_len += 1
                if qp is None:  # deletion
                    delt += 1
                elif qseq[qp] == rb:  # correct
                    correct += 1
                elif qseq[qp] != rb:  # sub
                    sub += 1
            else:  # we have an insertion
                ins += 1

    name = read.qname
    match = correct + sub
    length = match + ins + delt
    iden = 100 * float(match - sub) / match
    acc = 100 - 100 * float(sub + ins + delt) / length

    read_length = read.infer_read_length()
    masked_query_alignment_length = correct + sub + ins
    coverage = 100*float(masked_query_alignment_length) / read_length
    direction = '-' if read.is_reverse else '+'

    results = {
        "name": name,
        "coverage": coverage,
        "direction": direction,
        "length": length,
        "read_length": read_length,
        "match": match,
        "ins": ins,
        "del": delt,
        "sub": sub,
        "iden": iden,
        "acc": acc,
        "qstart": read.query_alignment_start,
        "qend": read.query_alignment_end,
        "rstart": read.reference_start,
        "rend": read.reference_end,
        "ref": references[read.reference_id],
        "aligned_ref_len": aligned_ref_len,
        "ref_coverage": 100 * float(aligned_ref_len) / lengths[read.reference_id],
        "masked": masked,
    }
    return results


def _process_reads(bam_fp, start_stop, all_alignments=False, min_length=None, bed_file=None):
    start, stop = start_stop
    counts = Counter()
    results = []
    with pysam.AlignmentFile(bam_fp, 'rb') as bam:
        if bed_file is not None:
            trees = intervaltrees_from_bed(bed_file)
        for i, read in enumerate(bam):
            if i < start:
                continue
            elif i == stop:
                break
            if read.is_unmapped:
                counts['unmapped'] += 1
                continue
            if not all_alignments and (read.is_secondary or read.is_supplementary):
                continue

            if bed_file is not None:
                if not trees[read.reference_name].overlaps(read.reference_start, read.reference_end):
                    sys.stderr.write('read {} does not overlap with any regions in bedfile\n'.format(read.query_name))
                    counts['masked'] += 1
                    continue
                else:
                    result = masked_stats_from_aligned_read(read, bam.references, bam.lengths, trees[read.reference_name])
            else:
                result = stats_from_aligned_read(read, bam.references, bam.lengths)

            if min_length is not None and result['length'] < min_length:
                counts['short'] += 1
                continue

            results.append(result)

    return counts, results


def main(arguments=None):
    args = parser.parse_args(arguments)

    if args.threads > 1 and args.bed is not None:
        pool = ProcessPoolExecutor(args.threads)
        mapper = pool.map
    else:
        pool = None
        mapper = map
        args.threads = 1

    headers = ['name', 'ref', 'coverage', 'ref_coverage', 'qstart', 'qend',
               'rstart', 'rend', 'aligned_ref_len', 'direction', 'length',
               'read_length', 'match', 'ins', 'del', 'sub', 'iden', 'acc']

    masked_headers = ['masked']
    if args.bed is not None:
        headers.extend(masked_headers)

    args.output.write('\t'.join(headers))
    args.output.write('\n')

    # create a slice of reads to process in each thread to avoid looping through
    # bam n read times and reduce mp overhead
    with pysam.AlignmentFile(args.bam) as bam:
        n_reads = bam.count()
    n_reads_per_proc = ceil(n_reads / args.threads)
    ranges = [(start, min(start + n_reads_per_proc, n_reads))
               for start in range(0, n_reads, n_reads_per_proc)]

    func = functools.partial(_process_reads, args.bam,
                             all_alignments=args.all_alignments,
                             min_length=args.min_length, bed_file=args.bed)

    counts = Counter()
    for batch_counts, results in mapper(func, ranges):
        counts.update(batch_counts)
        counts['total'] += len(results)
        for result in results:
            out_row = (str(result[x]) for x in headers)
            args.output.write('\t'.join(out_row))
            args.output.write('\n')
    if pool is not None:
        pool.shutdown(wait=True)

    if counts['total'] == 0:
        raise ValueError('No alignments processed. Check your bam and filtering options.')

    args.summary.write('Mapped/Unmapped/Short/Masked: {total}/{unmapped}/{short}/{masked}\n'.format_map(counts))


if __name__ == '__main__':
    main()



from __future__ import print_function
import random
import struct
import sys
import time
import os

import zmq

from msgpack import ExtType, packb, unpackb

class AlignClient(object):
    # A synchronous Python2 alignment client 

    REQ_PREFIX = struct.Struct('=HH')
    REQ_SUFFIX = struct.Struct('=Ld')
    RESP = struct.Struct('=HHLd?')

    def __init__(self, port):
        self.port = port
        self.ctx = zmq.Context()
        self.sock = self.ctx.socket(zmq.DEALER)

        self.sock.connect('tcp://127.0.0.1:{}'.format(self.port))

        self.counter = 0
        self.prefix = self.REQ_PREFIX.pack(os.getpid() % 0x10000,
                                           random.randrange(0x10000))

    def packb(self, data):
        return packb(data, encoding='utf-8', use_bin_type=True)

    def unpackb(self, packed):
        return unpackb(packed, use_list=False, encoding='utf-8')

    def _new_id(self):
        self.counter += 1
        if self.counter > 0xffffffff:
            self.counter = 0
        return (self.prefix + self.REQ_SUFFIX.pack(self.counter, time.time()),
                self.counter)

    def align(self, sequence):
        header, req_id = self._new_id()
        bname = 'align'.encode('utf-8')
        bargs = self.packb([sequence.decode()])
        bkwargs = self.packb(dict())
        msg = [header, bname, bargs, bkwargs]
        self.sock.send_multipart(msg)
        while True:
            try:
                data = self.sock.recv_multipart(zmq.NOBLOCK)
            except zmq.ZMQError as e:
                time.sleep(0.1)
            else:
                header, banswer = data
                pid, rnd, res_req_id, timestamp, is_error = self.RESP.unpack(header)
                if res_req_id != req_id:
                    raise ValueError('Received response for request {}, but send {}.'.format(res_res_id, req_id))
                answer = self.unpackb(banswer)
                return answer


if __name__ == '__main__':
    port, seq = sys.argv[1:3]
    client = AlignClient(port)
    alignments = client.align(seq)
    print(alignments)
            
